/*==============================================================
  Randsig.c  -> A system to generate random .signature files.
   v 1.5        (c) 1999 Santiago Romero AkA NoP / Compiler
                       sromero@escomposlinux.org

 -d option contributed by Grzegorz Adam Hankiewicz (gradha@iname.com)

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  Please read the README.* files to know how randsing works.
 =============================================================*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/stat.h>

#define START_CHARACTER1 '['
#define START_CHARACTER2 '*'
#define END_CHARACTER1   '*'
#define END_CHARACTER2   ']'
#define TAG_CHARACTER    't'
#define EXTTAG_CHARACTER 'e'
#define NUM_LINES         20

int CountTagsOnFile( FILE * );
void GetTagFromFile( FILE *, int, char * );
void ProcessFile( FILE *, FILE *, FILE *, int, char );
void RotateFile( FILE *, FILE *, FILE *, char *, int, char );
void TestFile( FILE *, FILE *, FILE *, int );
int SigDateDifferent( char *file );
void GetTagFromExternal( char *where );
void ShowFile( char, FILE *, char *, int, char );

char Ayuda[] = "\n Randsig v1.4 -> A program to generate random .signature files.\n\
                 1999  -  Santiago Romero  AkA  NoP / Compiler.\n\n\
 Program usage:       randsig [options]\n\n\
   OPTIONS:\n\
       --help              This help.\n\
       --verbose or -v     Write signature on the screen too.\n\
       --test or -t        Write $HOME/randsig.test file with all\n\
                           possible combinations to see results.\n\
       --rotate or -r      Instead of getting random signatures for\n\
                           each login, it gets them from the tag file\n\
                           in sequential order (starting with #1).\n\
       --daily or -d       Change signature only if it's last\n\
                           modification time is not today.\n\
       --show or -s        Show just a funny sentence.\n\
       Please create .randsig.external for use external tags.\n\n\
 Please read README.english or README.castellano files...\n";

  char *home;

/*--- Main ---------------------------------------------------*/
int main( int argc, char *argv[] )
{
  FILE *main, *tags, *out;
  int numtags, loop_arg;
  char fichero[128], currenttag[128];
  char verbose = 0, rotate = 0, test = 0, daily = 0, show=0;

   for (loop_arg = 1; loop_arg < argc; loop_arg++)
   {
      if (strstr( (const char *) argv[loop_arg], "--help") != NULL ||
         strstr( (const char *) argv[loop_arg], "-help") != NULL ||
         strstr( (const char *) argv[loop_arg], "-?") != NULL ||
         strstr( (const char *) argv[loop_arg], "/?") != NULL ||
         strstr( (const char *) argv[loop_arg], "-h") != NULL )
      {
         printf ("%s\n", Ayuda);
         return 0;
      }

      if( strstr( (const char *) argv[loop_arg], "--verbose") != NULL ||
         strstr( (const char *) argv[loop_arg], "-v") != NULL )
      {
         verbose = 1;
      }

      if( strstr( (const char *) argv[loop_arg], "--rotate") != NULL ||
         strstr( (const char *) argv[loop_arg], "-r") != NULL )
      {
         rotate = 1;
      }

      if( strstr( (const char *) argv[loop_arg], "--daily") != NULL ||
         strstr( (const char *) argv[loop_arg], "-d") != NULL )
      {
         daily = 1;
      }

     if( strstr( (const char *) argv[loop_arg], "--test") != NULL ||
       strstr( (const char *) argv[loop_arg], "-t") != NULL)
      {
         test = 1;
      }
     if( strstr( (const char *) argv[loop_arg], "--show") != NULL ||
       strstr( (const char *) argv[loop_arg], "-s") != NULL)
      {
         show = 1;
      }
   }

  home = getenv("HOME");
  srandom(time(NULL));

  sprintf(currenttag, "%s/.randsig.rotate", home );
  sprintf(fichero, "%s/.signature.main", home );
  main = fopen( fichero, "rt" );
  if( main == NULL )
  {
     printf("\n$HOME/.signature.main file not found.\n");
     printf("\nPlease read the RandSig README file.\n\n");
     return(-1);
  }

  sprintf(fichero, "%s/.signature.tags", home );
  tags = fopen( fichero, "rt" );
  if( tags == NULL )
  {
     printf("\n$HOME/.signature.tags file not found.\n");
     printf("\nPlease read the RandSig README file.\n\n");
     return(-1);
  }

  sprintf(fichero, "%s/.signature", home );
  if (daily && !SigDateDifferent(fichero))
  {
      fclose(main);
      fclose(tags);
      /* The file has today's date. Don't update */
      return 1;
  }

  numtags=CountTagsOnFile( tags );

  if( !test )
  {
    if( !show )
    {
       out = fopen( fichero, "wt" );
       if( !rotate )
          ProcessFile( main, tags, out, numtags, verbose );
       else
          RotateFile( main, tags, out, currenttag, numtags, verbose );
    }
    else
          ShowFile( rotate, tags, currenttag, numtags, verbose );

  }
  else
  {
       sprintf(fichero, "%s/randsig.test", home );
       out = fopen( fichero, "wt" );
       TestFile( main, tags, out, numtags );
  }

  fcloseall();
  return(0);

}


/*--------------------------------------------------------------
  This function process the .signature.main file writing all
  it's characters to .signature, except the control codes
  such as [t, replaced by the right command.
 -------------------------------------------------------------*/
void ProcessFile( FILE *main, FILE * tags,
                 FILE * out, int numtags, char verbose )
{
   char last, current = 0;
   char tagtext[80*NUM_LINES];
   int tag;

   rewind(main);
   /* start from the beginning of the file: */
   while( !feof(main) )
   {
      last = current;
      current = fgetc(main);

      if( current != START_CHARACTER1 && last != START_CHARACTER1 )
      {
        if( current != EOF )
        {
         fputc( current, out );
         if( verbose ) fputc( current, stdout );
        }
      }

      /* do we have found a new tag? */
      if( last == START_CHARACTER1 && current == TAG_CHARACTER  )
      {
         tag = (random()%numtags)+1;
         GetTagFromFile( tags, tag, tagtext );
         fputs( tagtext, out );
         if( verbose ) fputs( tagtext, stdout );
      }

      /* do we have found a new EXTERNAL tag? */
      else if( last == START_CHARACTER1 && current == EXTTAG_CHARACTER  )
      {
         GetTagFromExternal( tagtext );
         fputs( tagtext, out );
         if( verbose ) fputs( tagtext, stdout );
      }

      else if( last == START_CHARACTER1 && current != TAG_CHARACTER )
      {
        fputc(START_CHARACTER1, out );
        if( verbose ) fputc( START_CHARACTER1, stdout );
        if( current != EOF )
        {
           fputc(current, out );
           if( verbose ) fputc( current, stdout );
        }
      }
   }
   if( verbose ) fprintf( stdout, "\n");
}


/*--------------------------------------------------------------
  This function process the .signature.main file writing all
  it's characters to .signature, except the control codes
  such as [t, replaced by the right command.
 -------------------------------------------------------------*/
void ShowFile( char rotate, FILE * tags, char *numtag, int numtags, char verbose )
{
   char last, current = 0;
   char tagtext[80*NUM_LINES];
   FILE *filenumtag;
   int tag;

   if(!rotate)
   {
      tag = (random()%numtags)+1;
      GetTagFromFile( tags, tag, tagtext );
      fputs( tagtext, stdout );
      fprintf( stdout, "\n");
      return;
   }
   else
   {
      filenumtag = fopen(numtag,"rt");
      if( filenumtag == NULL )
      {
         filenumtag = fopen(numtag,"wt");
         fprintf(filenumtag, "%d\n", 0 );
         fclose(filenumtag);
         filenumtag = fopen(numtag,"rt");
         if( filenumtag == NULL )
         {
            printf("Error creating file %s... :?\n");
            exit(-1);
         }
      }
      rewind(filenumtag);
      fscanf(filenumtag, "%d", &tag );
      if( tag+1 < numtags+1 ) tag++;
      else                  tag=1;
      GetTagFromFile( tags, tag, tagtext );
      fputs( tagtext, stdout );
      fprintf( stdout, "\n");
      fclose(filenumtag);
      filenumtag = fopen(numtag,"wt");
      fprintf(filenumtag, "%d\n", tag );
      fclose(filenumtag);
   }
}


/*--------------------------------------------------------------
  This function process the .signature.main file writing all
  it's characters to .signature, except the control codes
  such as [t, replaced by the right command. It reads tags
  sequentially instead of randomly.
 -------------------------------------------------------------*/
void RotateFile( FILE *main, FILE * tags, FILE * out,
                 char *numtag, int numtags, char verbose )
{
   char last, current = 0;
   char tagtext[80*NUM_LINES];
   int tag;
   FILE *filenumtag;

   filenumtag = fopen(numtag,"rt");
   if( filenumtag == NULL )
   {
      filenumtag = fopen(numtag,"wt");
      fprintf(filenumtag, "%d\n", 0 );
      fclose(filenumtag);
      filenumtag = fopen(numtag,"rt");
      if( filenumtag == NULL )
      {
         printf("Error creating file %s... :?\n");
         exit(-1);
      }
   }

   rewind(filenumtag);
   rewind(main);


   /* start from the beginning of the file: */
   while( !feof(main) )
   {
      last = current;
      current = fgetc(main);

      if( current != START_CHARACTER1 && last != START_CHARACTER1 )
      {
        if( current != EOF )
        {
          fputc( current, out );
          if( verbose ) fputc( current, stdout );
        }
      }

      /* do we have found a new tag? */
      if( last == START_CHARACTER1 && current == TAG_CHARACTER  )
      {

         fscanf(filenumtag, "%d", &tag );
         if( tag+1 < numtags+1 ) tag++;
         else                  tag=1;
         GetTagFromFile( tags, tag, tagtext );
         fputs( tagtext, out );
         if( verbose ) fputs( tagtext, stdout );

         fclose(filenumtag);
         filenumtag = fopen(numtag,"wt");
         fprintf(filenumtag, "%d\n", tag );
         fclose(filenumtag);
      }
      /* do we have found a new EXTERNAL tag? */
      else if( last == START_CHARACTER1 && current == EXTTAG_CHARACTER  )
      {
         GetTagFromExternal( tagtext );
         fputs( tagtext, out );
         if( verbose ) fputs( tagtext, stdout );
      }
      else if( last == START_CHARACTER1 && current != TAG_CHARACTER )
      {
        fputc(START_CHARACTER1, out );
        if( verbose ) fputc( START_CHARACTER1, stdout );
        if( current != EOF )
        {
           fputc(current, out );
           if( verbose ) fputc( current, stdout );
        }
      }
   }
   if( verbose ) fprintf( stdout, "\n");
}



/*--------------------------------------------------------------
  This function writes a randsig.test file.
 -------------------------------------------------------------*/
void TestFile( FILE *main, FILE * tags, FILE * out, int numtags )
{
   char last, current = 0;
   char tagtext[80*NUM_LINES];
   int tag, i = 0;

   for( i=1; i<=numtags; i++ )
   {
     rewind(main);

     while( !feof(main) )
     {
        last = current;
        current = fgetc(main);

        if( current != START_CHARACTER1 && last != START_CHARACTER1 )
        {
          if( current != EOF )
           fputc( current, out );
        }

        if( last == START_CHARACTER1 && current == TAG_CHARACTER  )
        {
           tag = i;
           GetTagFromFile( tags, tag, tagtext );
           fputs( tagtext, out );
        }
        /* do we have found a new EXTERNAL tag? */
        else if( last == START_CHARACTER1 && current == EXTTAG_CHARACTER  )
        {
          GetTagFromExternal( tagtext );
          fputs( tagtext, out );
        }
        else if( last == START_CHARACTER1 && current != TAG_CHARACTER )
        {
          fputc(START_CHARACTER1, out );
          if( current != EOF )
             fputc(current, out );
        }

     }
     fprintf(out, "\n\n");
   }


}



/*--------------------------------------------------------------
  int CountTagsOnFile( tagfile );
  This function just counts the number of taglines available
  on the specified file (usually .signature.tags) by simply
  reading all the file and looking for all the "[*" strings.
 -------------------------------------------------------------*/
int CountTagsOnFile( FILE * fp )
{
   char last, current = 0;
   int numtags = 0;

   rewind(fp);
   /* start from the beginning of the file: */
   while( !feof(fp) )
   {
      last = current;
      current = fgetc(fp);

      /* do we have found a new tag? */
      if( last == START_CHARACTER1 &&
          current == START_CHARACTER2 )
         numtags++;
   }
   return(numtags);
}


/*--------------------------------------------------------------
  int SigDateDifferent( sigfile );
  This functions checks the actual date and the date of the
  .signature file. If the file doesn't exist or it's date is
  not the same as today, return 1. Otherwise return zero.
 -------------------------------------------------------------*/
int SigDateDifferent( char *file )
{
   struct stat info;
   struct tm filetime, *t;
   time_t daytime;

   FILE *sigfile = fopen (file, "rt");
   if (!sigfile)
      return 1; /* file doesn't exist. */
   fclose (sigfile);

   if (stat (file, &info))
      return 0; /* error getting file data */

   t = localtime (&info.st_mtime);
   memcpy (&filetime, t, sizeof (struct tm));

   time (&daytime);
   t = localtime (&daytime);

   if (t->tm_mday == filetime.tm_mday && t->tm_mon == filetime.tm_mon &&
      t->tm_year == filetime.tm_year)
   {
      /* File shouldn't be updated */
      return 0;
   }
   return 1;
}


/*--------------------------------------------------------------
  void GetTagFromFile( tagfile, numtag, string );
  This function reads another time the tag file searching for
  the nth tag located on it (and passing though the n-1 ones).
 -------------------------------------------------------------*/
void GetTagFromFile( FILE *fp, int tag, char *where )
{

   char last, current = 0;
   int numtags = 0, index=0;

   rewind(fp);
   while( !feof(fp) )
   {
      last = current;
      current = fgetc(fp);
      if( last == START_CHARACTER1 &&
          current == START_CHARACTER2 )
      {
         numtags++;

         /* if this is our tag, extract it */
         if( numtags == tag )
         {
              last = current = 0;
              while( last != END_CHARACTER1 &&
                     current != END_CHARACTER2 )
              {
                 current = last;
                 current = fgetc(fp);
                 where[index] = current;
                 index++;
              }
              where[index-2] = '\0';
              return;
         }
      }
   }

}

/*--------------------------------------------------------------
  void GetTagFromExternal( string );
  Gets the output of an external tagline program,
  such as fortune. Fortune adds an extra \n, so... :)
 -------------------------------------------------------------*/
void GetTagFromExternal( char *where )
{
   FILE *external;
   char fname[256];

   sprintf(fname, "%s/.randsig.external", home );
   external = popen( fname, "r" );
   if( external == NULL )
   {
      strcpy( where, "Cannot execute %s script: please create it\nas a script calling the external program you want to use.\nPlease read /usr/doc/randsig/README.\n");
      return;
   }
   memset( where, 0, (80*NUM_LINES)-1 );
   fread( where, 80*NUM_LINES, 1, external );
   pclose(external);
}
